/* Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : OUTPUT_MUX_TRI

//`define OUTPUT_MUX_TRI

// Module Name: rtavr_ior_usart
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
// SUPPORTED FORMAT (TX): 
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | 1  |                   7N1
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | 1  | 1  |              7N2
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | P  | 1  |              7P1
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7 | 1  |              8N1
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | P  | 1  | 1  |         7P2
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7 | 1  | 1  |         8P2
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7 | P  | 1  |         8P1
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7 | D8 | 1  |         9N1
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7 | P  | 1  | 1  |    8P2
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7 | D8 | 1  | 1  |    9N2
//
//
//
//                          DO:  NORMAL    TRISTATE(FAKE)   
// Number of Slice Flip Flops:     16         16
// Number of 4 input LUTs:         22         23
// Number of occupied Slices:      19         20
// Total Number of 4 input LUTs:   22         23 
// Number of bonded IOBs :         33         33
//   IOB Flip Flops :              16         16
// Number of BUFGMUXs :             1          1
// Number of RAMB16BWEs :          -           -

module rtavr_ior_usart(
		input CLK,
		input RESET,

		input [7:0] DI,
		output [7:0] DO
		, input WR,RD
		, input [5:0] ADDR
		// external I/O
		, output TX_OUT
		, input RX_IN
		, input XCK_IN
		, input XCK_DDR
		, output XCK_OUT

		, output TXEN
		, output XCKEN
		// interrupt
		, output TX_INT
		, output UDRE_INT
		, output RX_INT
		, input RST_TX_INT
		, input RST_UDRE_INT
		, input RST_RX_INT
		);
  parameter BASE_ADDR = 6'h0e;  // USART : (0x0e - 0x14 ADC space)
				//  UBRRL, UBRRH, UCSRC, UCSRB, UCSRA, UDR

  reg  [11:0] r_ps;
  reg  tx_mclk;

  reg  [3:0] tx_ps;
  reg  [4:0] tx_count;
  reg  [1:0] r_count_max_x;
  wire [4:0] v_count_max = { 2'b10, r_count_max_x , 1'b1};
  reg  tx_count_en;
  reg  [9:0] tx_data;
  reg  tx_data_load;
  reg  r_txd;

  reg  [4:0] rx_ps;
  reg  [4:0] rx_count;
  reg  rx_count_en;
  reg  [10:0] rx_data;
  reg  rx_data_rdy;
  reg  rx_parity;
  reg  rx_c;

  reg  [7:0] r_udr_tx;
  reg  [7:0] r_udr_rx;
  reg  [11:0] UBRR;

  reg  RXC ;
  reg  TXC;
  wire UDRE;
  reg  FE ;
  reg  DOR;
  reg  UPE ;
  reg  U2X;
  reg  MPCM;

  reg  RXCIE;
  reg  TXCIE;
  reg  UDRIE;
  reg  rx_en ;
  reg  tx_en;
  reg  RXB8;
  reg  TXB8;

  reg  [1:0] UMSEL;
  reg  [1:0] UPM;
  reg  USBS;
  reg  r_ucsz2;
  reg  r_ucsz0;
//  wire UCSZ = { r_ucsz2, 1'b1, r_ucsz0 };
  reg  UCPOL;

  wire OE_ubrrl = RD & (ADDR == BASE_ADDR);
  wire WE_ubrrl = WR & (ADDR == BASE_ADDR);
  wire OE_ubrrh = RD & (ADDR == (BASE_ADDR + 1)); 
  wire WE_ubrrh = WR & (ADDR == (BASE_ADDR + 1));
  wire OE_ucsrc = RD & (ADDR == (BASE_ADDR + 2)); 
  wire WE_ucsrc = WR & (ADDR == (BASE_ADDR + 2));
  wire OE_ucsrb = RD & (ADDR == (BASE_ADDR + 3)); 
  wire WE_ucsrb = WR & (ADDR == (BASE_ADDR + 3));
  wire OE_ucsra = RD & (ADDR == (BASE_ADDR + 4)); 
  wire WE_ucsra = WR & (ADDR == (BASE_ADDR + 4));
  wire OE_udr   = RD & (ADDR == (BASE_ADDR + 5)); 
  wire WE_udr   = WR & (ADDR == (BASE_ADDR + 5));

`ifdef OUTPUT_MUX_OR
  wire [7:0] i_DO = OE_ubrrl ? UBRR[7:0] 
		: OE_ubrrh ? { 4'h0 , UBRR[11:8] }
		: OE_ucsra ? { RXC, TXC, UDRE, FE, DOR, UPE, U2X, MPCM}
		: OE_ucsrb ?  {RXCIE, TXCIE, UDRIE, rx_en,
				  tx_en, r_ucsz2 , RXB8, TXB8 }
		: OE_ucsrc ? { UMSEL, UPM[1:0], USBS, 1'b1 , r_ucsz0, UCPOL}
		: OE_udr ? r_udr_rx : 0;
`else
  wire [7:0] i_DO = OE_ubrrl ? UBRR[7:0] 
		: OE_ubrrh ? { 4'h0 , UBRR[11:8] }
		: OE_ucsra ? { RXC, TXC, UDRE, FE, DOR, UPE, U2X, MPCM}
		: OE_ucsrb ?  {RXCIE, TXCIE, UDRIE, rx_en,
				  tx_en, r_ucsz2 , RXB8, TXB8 }
		: OE_ucsrc ? { UMSEL, UPM[1:0], USBS, 1'b1, r_ucsz0, UCPOL}
		: r_udr_rx;
`endif

`ifdef OUTPUT_MUX_TRI
  bufif1(DO, i_DO, (OE_ubrrl | OE_ubrrh | OE_ucsrb| OE_ucsra | OE_udr));
`else
  assign DO = i_DO;
`endif

  assign UDRE	  = ~(tx_data_load);
  assign TX_INT   = TXC  & TXCIE;
  assign RX_INT   = RXC  & RXCIE;
  assign UDRE_INT = UDRE & UDRIE;
  assign TX_OUT   = UMSEL[1] ? tx_data[0] : r_txd;
  assign XCK_OUT  = tx_mclk;
  assign XCKEN    = (tx_en | rx_en) & UMSEL[0];
  assign TXEN	  = tx_en;

  // v_tclk/r_tclk  : internal clock for tx
  wire v_tclk = (UMSEL[0] & ~UCPOL & ~XCK_DDR ) ? XCK_IN 
  		 : (UMSEL[0] & UCPOL & ~XCK_DDR ) ? ~XCK_IN 
  		 : (UMSEL[1] & UMSEL[0] & XCK_DDR ) ? tx_ps[0]
		 : (U2X == 1) ? tx_ps[2]
		 : tx_ps[3] ;
  reg  r_tclk;
 
  wire do_txcount = tx_count_en & (r_tclk ^ v_tclk);
  wire do_txin  =  do_txcount &  r_tclk;
  wire do_txsft =  do_txcount & ~r_tclk;


  wire [1:0] v_bit_num_x = 
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | 1  |                   7N1
			   (~r_ucsz0 & ~UPM[1] & ~USBS) ? 0
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | 1  | 1  |              7N2
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | P  | 1  |              7P1
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7 | 1  |              8N1
  			 :  (~r_ucsz0 & (UPM[1] ^ USBS))
  			   |(r_ucsz0 & ~r_ucsz2 & ~UPM[1] & ~USBS) ? 1
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | P  | 1  | 1  |         7P2
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7 | 1  | 1  |         8P2
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7 | P  | 1  |         8P1
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7 | D8 | 1  |         9N1
  			 :  (~r_ucsz0 & UPM[1] & USBS)
  			   |(r_ucsz0 & ~r_ucsz2 & (UPM[1] ^ USBS))
  			   |(r_ucsz2 & ~UPM[1] & ~USBS) ? 2
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7 | P  | 1  | 1  |    8P2
// | 0 | D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7 | D8 | 1  | 1  |    9N2
  			 :  3 ;

  wire tx_parity7 = r_udr_tx[0] ^ r_udr_tx[1] ^ r_udr_tx[2] ^ r_udr_tx[3]
		   ^ r_udr_tx[4] ^ r_udr_tx[5] ^ r_udr_tx[6];
  wire tx_parity8 = tx_parity7 ^ r_udr_tx[7];


`ifdef USART_LOCAL_LB
  wire v_rxd = r_txd;
`else
  wire v_rxd = RX_IN;
`endif

  // v_rclk/r_rclk  : internal clock for rx
  wire v_rclk = (UMSEL[0] & ~UCPOL & ~XCK_DDR ) ? XCK_IN 
  		 : (UMSEL[0] & UCPOL & ~XCK_DDR ) ? ~XCK_IN 
  		 : (UMSEL[1] & UMSEL[0] & XCK_DDR ) ? rx_ps[0]
		 : (U2X == 1) ? rx_ps[2]
		 : rx_ps[3] ;
  reg  r_rclk;

  wire do_rx = (r_rclk ^ v_rclk);
  wire do_rxcount = do_rx & rx_count_en;
  wire do_rxin  =  do_rxcount &  r_rclk;
  wire do_rxsft =  do_rxcount & ~r_rclk;

//  wire cond_udre_1 = ~WE_spdr & (tx_count == v_count_max);

  always @(negedge CLK)
  begin
    if (RESET == 0)
      begin
      end
    else
      begin
      end
  end

  always @(posedge CLK)
  begin
    if (RESET == 0)
      begin
	r_ps <= 0;
	r_tclk <= 0;
	r_rclk <= 0;
	tx_mclk <= 0;

	tx_ps <= 0;
	tx_data_load <= 0;
	tx_count <= 0;
	r_count_max_x <= 0;
	tx_count_en <= 0;
	tx_data <= 0;
	r_txd <= 1;

	rx_ps <= 0;
	rx_count <= 0;
	rx_count_en <= 0;
	rx_data <= 0;
	rx_data_rdy <= 0;

	UBRR <= 0;
	r_udr_rx <= 0;
	r_udr_tx <= 0;

	RXC  <= 0;
	TXC <= 0;
	FE  <= 0;
	DOR <= 0;
	UPE  <= 0;
	U2X <= 0;
	MPCM <= 0;

	RXCIE  <= 0;
	TXCIE  <= 0;
	UDRIE  <= 0;
	rx_en  <= 0;
	tx_en  <= 0;
	RXB8 <= 0;
	TXB8 <= 0;

	UMSEL <= 0;
	UPM <= 0;
	USBS <= 0;
	r_ucsz0 <= 0;
	r_ucsz2 <= 0;
	UCPOL <= 0;
      end
    else
      begin
	r_tclk <= v_tclk;
	r_rclk <= v_rclk;

	if (r_ps == UBRR) r_ps <= 0;
	else r_ps <= r_ps + 1;

	if(WE_udr)
	  begin
	    r_udr_tx <= DI;
	    tx_data_load <= 1;
	  end

	if (tx_data_load & ~tx_count_en & ~r_tclk)
	  begin
	    tx_data_load <= 0;
	    tx_ps <= 0;
	    tx_count <= UMSEL[1] ? 5'b00010 : 5'b11110;
	    tx_count_en <= 1;
	    r_count_max_x <= v_bit_num_x;

	    tx_data[0] <= 0; // start bit
	    tx_data[7:1] <= r_udr_tx[6:0]; 
	    tx_data[8] <= (~r_ucsz0 & UPM[1]) ? (UPM[0] ^ tx_parity7)
	    		 : ( r_ucsz0 ) ? r_udr_tx[7]
			 : 1;
	    tx_data[9] <= (r_ucsz0 & ~r_ucsz2 & UPM[1]) ? (UPM[0] ^ tx_parity8)
			 :(r_ucsz0 & r_ucsz2) ? TXB8
			 : 1;
	  end
        else if (tx_count_en & (r_ps == UBRR))
	   tx_ps <= tx_ps + 1;

	if (do_txcount)
	  begin
	    tx_count <= tx_count + 1;
	    if (tx_count == v_count_max) tx_count_en <= 0;
	  end
	if (do_txsft)
	  begin
	    r_txd <= tx_data[0];
	    tx_data[8:0] <= tx_data[9:1];
	    tx_data[9]   <= 1;
	    tx_mclk <= 0;
	  end
	if (do_txin)
	  begin
	    tx_mclk <= 1;
	  end

//	if(RST_INT_UDRE | (WE_spdr & OE_spdr) ) r_spif <= 0;
//	else if (cond_spif_1 | cond_spif_2) r_spif <= 1;

	if (RST_TX_INT | WE_ucsra & DI[6]) TXC <= 0;
        else if ( do_txcount & (tx_count == v_count_max) & !tx_data_load)
		 TXC <= 1;
// 	else if (~(tx_data_load | tx_count_en)) TXC <= 1;


	if (do_rx & ~rx_count_en & ~r_rclk & ~v_rxd)
	  begin
	    rx_parity <= 0;
	    r_ps <= 0;
//	    rx_count <= 5'b11110;
//	    rx_count <= 5'b00000;
	    rx_count <= 5'b00010;
	    rx_count_en <= 1;
	    r_count_max_x <= v_bit_num_x;
	  end
        else if (r_ps == UBRR)
	   rx_ps <= rx_ps + 1;

	if (do_rxcount)
	  begin
	    rx_count <= rx_count + 1;
	    if (rx_count == v_count_max) rx_count_en <= 0;
	    if (rx_count == v_count_max) rx_data_rdy <= 1;
	  end
	if (do_rxin)
	  begin
	    rx_c <= v_rxd;
	    rx_parity <= rx_parity ^ v_rxd;
	  end
	if (do_rxsft)
	  begin
	    rx_data[6:0] <= rx_data[7:1];
	    rx_data[7] <= (v_bit_num_x == 0) ? rx_c : rx_data[8];
	    rx_data[8] <= (v_bit_num_x == 1) ? rx_c : rx_data[9];
	    rx_data[9] <= (v_bit_num_x == 2) ? rx_c : rx_data[10];
	    rx_data[10] <= (v_bit_num_x == 3) ? rx_c : 0;
	    //rx_data[10] <= rx_c;
	  end

	if (rx_data_rdy & ~rx_count_en & ~r_rclk)
	  begin
	    rx_data_rdy <= 0;
	    r_udr_rx[7:0] <= rx_data[7:0];
	    RXB8 <= rx_data[8];
	  end

	if (RST_RX_INT | OE_udr) RXC   <= 0;
	else if (rx_data_rdy & ~rx_count_en & ~r_rclk) RXC <= 1;

	if(WE_ubrrl) UBRR[7:0] <= DI;
	if(WE_ubrrh) UBRR[11:8] <= DI[3:0];

	//if(RST_INT_UDRE | 0) UDRE  <= 0;
	if(WE_ucsra) FE    <= DI[4];
	if(WE_ucsra) DOR   <= DI[3];
	if(WE_ucsra) UPE    <= DI[2];
	if(WE_ucsra) U2X   <= DI[1];
	if(WE_ucsra) MPCM  <= DI[0];


	if(WE_ucsrb) RXCIE   <= DI[7];
	if(WE_ucsrb) TXCIE   <= DI[6];
	if(WE_ucsrb) UDRIE   <= DI[5];
	if(WE_ucsrb) rx_en    <= DI[4];
	if(WE_ucsrb) tx_en    <= DI[3];
	if(WE_ucsrb) RXB8    <= DI[1];
	if(WE_ucsrb) TXB8    <= DI[0];

	if(WE_ucsrc) UMSEL   <= DI[7:6];
	if(WE_ucsrc) UPM     <= DI[5:4];
	if(WE_ucsrc) USBS    <= DI[3];
	if(WE_ucsrc) UCPOL   <= DI[0];

	if(WE_ucsrc) r_ucsz0  <= DI[1];
	if(WE_ucsrb) r_ucsz2  <= DI[2];

      end
  end

endmodule
