/* Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : OUTPUT_MUX_TRI

//`define OUTPUT_MUX_TRI

// Module Name: rtavr_ior_spi
//  Version 003
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                               NORMAL  (1)  (2)  (3)  (4)   (5)   (6)  (7)
// Number of Slice Flip Flops:     39    31   37   37   37    30    28   28
// Number of 4 input LUTs:        119    54   70   91   93    49    47   46
// Number of occupied Slices:      78    44   52   62   63    41    39   34
// Total Number of 4 input LUTs:  126    62   78   97   99    57    55   46
// Number of bonded IOBs :         38    36   36   36   36    36    35   35
//   IOB Flip Flops :               - 
// Number of BUFGMUXs :             1  
// Number of RAMB16BWEs :           - 
//
// (1) FIXED_MSTR=1, CPOL=0, CPHA=0, SPR=0
// (2) FIXED_MSTR=1, CPOL=0, CPHA=0
// (3) FIXED_MSTR=1, CPOL=0
// (4) FIXED_MSTR=1
// (5)  FIXED_MSTR=1, CPOL=0, CPHA=0, SPR=0 , NO_WCOL
// (6)  FIXED_MSTR=1, CPOL=0, CPHA=0, SPR=0 , NO_WCOL , NO_SS_INT

// 
// i_sck : SCK_IN ^ CPOL
//          ___     ___     ___     ___     ___     ___     ___     ___
//  _______|   |___|   |___|   |___|   |___|   |___|   |___|   |___|   |____
// 
// CPHA=0 IN  SFT IN  SFT IN  SFT IN  SFT IN  SFT IN  SFT IN  SFT IN  SFT  
// CPHA=1 SFT IN  SFT IN  SFT IN  SFT IN  SFT IN  SFT IN  SFT IN  SFT IN
// 
// 
//     SFT :    +-----------------+
//              |                 |   DORD == 0 : ROL  , 1 : ROR
//              +--  C -- SPDR  --+ 
// 
//    IN       CPHA == 0              :  C 
//             CPHA == 1 & DORD == 0  :  SPDR[0]
//             CPHA == 1 & DORD == 1  :  SPDR[7]
//    
//    OUT      CPHA == 0 & DORD == 0  :  SPDR[7]
//             CPHA == 0 & DORD == 1  :  SPDR[0]
//  

module rtavr_ior_spi(
		input CLK,
		input RESET,

		input [7:0] DI,
		output [7:0] DO
		, input WR,RD
		, input [5:0] ADDR
		// external I/O
		, input SS_DDR
		, input SS_IN
		, input SCK_IN
		, input MOSI_IN
		, input MISO_IN
		, output SPI_MASTER
		, output SPI_SLAVE
		, output SCK_OUT
		, output MOSI_OUT
		, output MISO_OUT
		// interrupt
		, output SPI_INT
		, input RST_SPI_INT
		);
  parameter BASE_ADDR  = 6'h2e;    // SPI : SPDR 0x2e, SPSR, SPCR

  reg  [6:0] r_ps;
  reg  r_count_en;
  reg  [3:0] r_count;
  reg  r_ss;
  reg  r_ss_prev;
  reg  r_spsr_lck;
  reg  r_spi_mclk;
  reg  c;
  
  reg  [7:0] r_spdr;
  reg  [7:0] r_spcr;
  reg  r_spif; // spsr bit7
  reg  r_wcol; // spsr bit6
  reg  r_spi2x; // spsr bit0

  wire SPIE = r_spcr[7];
  wire SPE = r_spcr[6];
`ifdef FIXED_DORD
  wire DORD = `FIXED_DORD;
`else
  wire DORD = r_spcr[5];
`endif
`ifdef FIXED_MSTR
  wire MSTR = `FIXED_MSTR;
`else
  wire MSTR = r_spcr[4];
`endif
`ifdef FIXED_CPOL
  wire CPOL = `FIXED_CPOL;
`else
  wire CPOL = r_spcr[3];
`endif
`ifdef FIXED_CPHA
  wire CPHA = `FIXED_CPHA;
`else
  wire CPHA = r_spcr[2];
`endif
`ifdef FIXED_SPR
  wire [2:0] SPR = `FIXED_SPR;
`else
  wire [2:0] SPR = { r_spcr[1:0] , ~r_spi2x };
`endif

  wire OE_spdr = RD & (ADDR == BASE_ADDR);
  wire WE_spdr = WR & (ADDR == BASE_ADDR);
  wire OE_spsr = RD & (ADDR == (BASE_ADDR + 1)); 
  wire WE_spsr = WR & (ADDR == (BASE_ADDR + 1));
  wire OE_spcr = RD & (ADDR == (BASE_ADDR + 2)); 
  wire WE_spcr = WR & (ADDR == (BASE_ADDR + 2));

  wire [7:0] i_DO = OE_spdr ? r_spdr[7:0] 
		: OE_spsr ? { r_spif, r_wcol, 5'b0, r_spi2x }
		: OE_spcr ? r_spcr: 0;
`ifdef OUTPUT_MUX_TRI
  bufif1(DO, i_DO, (OE_spdr | OE_spsr | OE_spcr));
`else
  assign DO = i_DO;
`endif

  wire v_spi_mclk = (SPR == 0) ? r_ps[0]
  		  : (SPR == 1) ? r_ps[1]
  		  : (SPR == 2) ? r_ps[2]
  		  : (SPR == 3) ? r_ps[3]
  		  : (SPR == 4) ? r_ps[4]
  		  : (SPR == 5) | (SPR == 6) ? r_ps[5]
  		  : r_ps[6];

  wire v_spi_out = ~CPHA & ~DORD ? r_spdr[7]
	         : ~CPHA &  DORD ? r_spdr[0]
	         :  c;

  assign SPI_MASTER = SPE & MSTR;
  assign SPI_SLAVE = SPE & ~MSTR;
  assign SCK_OUT = r_spi_mclk;
  assign MOSI_OUT = v_spi_out;
  assign MISO_OUT = v_spi_out;
  assign SPI_INT = r_spif & SPIE;

  wire v_spi_in = MSTR ? MISO_IN : MOSI_IN;

  wire i_sck = (MSTR ? v_spi_mclk : (SCK_IN ^ CPOL) ) & ~SS_IN & SPE;
  reg  r_sck;

  wire do_count = (r_sck ^ i_sck);
  wire do_in  =  do_count & ~(r_sck ^ CPHA);
  wire do_sft =  do_count &  (r_sck ^ CPHA);


  wire cond_spif_1 = ~WE_spdr & (r_count == 15);
  wire cond_spif_2 = ~r_ss_prev & r_ss;
  wire cond_wcol   =  WE_spdr & ~(r_count == 0);

  always @(posedge CLK)
  begin
    if (~RESET | ~SPE)
      begin
	r_ss		<= 0;
	r_ss_prev	<= 0;
      end
    else
      begin
	r_ss		<= SPE & MSTR & ~SS_DDR & ~SS_IN;
	r_ss_prev	<= r_ss;
      end
  end

  always @(posedge CLK)
  begin

    if (RESET == 0)
      begin
	c	<= 0;
	r_ps	<= 0;
	r_spdr <= 0;
	r_spcr <= 0;
	r_spif <= 0;
	r_wcol <= 0;
	r_spsr_lck <= 0;
	r_spi2x <= 0;

  	r_sck <= 0;
	r_count_en <= 0;
	r_count <= 0;
	r_spi_mclk <= 0;
      end
    else
      begin
  	r_sck <= i_sck;
	
	if (WE_spdr)
	  begin
	    r_spdr <= DI;
	    r_count <= 0;
	    r_ps <= 0;
	    r_count_en <= 1;
	  end
	else 
	  begin
//	    if (r_count_en) r_ps <= r_ps + 1;
	    if (r_count_en & ~(do_count & (r_count == 15))) r_ps <= r_ps + 1;
	    if (do_count)
	      begin
		r_count <= r_count + 4'b0001;
		if (r_count == 15) r_count_en <= 0;
	      end
	    if (do_sft)
	      begin
		if (DORD) { r_spdr[7:0], c}  <= { c , r_spdr[7:0]};
		else      { c , r_spdr[7:0]} <= { r_spdr[7:0], c} ;
		r_spi_mclk <= (CPOL ^ CPHA);
	      end
	    if (do_in)
	      begin
		if (CPHA &  DORD) r_spdr[7] <= v_spi_in;
		if (CPHA & ~DORD) r_spdr[0] <= v_spi_in;
		if (~CPHA)                c <= v_spi_in;
		r_spi_mclk <= ~(CPOL ^ CPHA);
	      end
	  end

	if(OE_spsr & (r_spif | r_wcol)) r_spsr_lck <= 0;
	else if (cond_spif_1 | cond_spif_2 | cond_wcol) r_spsr_lck <= 1;

	if(RST_SPI_INT | ~r_spsr_lck & (WE_spdr | OE_spdr) ) r_spif <= 0;
	else if (cond_spif_1 | cond_spif_2) r_spif <= 1;

	if(~r_spsr_lck & (WE_spdr | OE_spdr) ) r_wcol <= 0;
	else if (cond_wcol) r_wcol <= 1;

	if(WE_spcr) r_spcr <= DI;
//	if(WE_spsr) r_spif <= DI[7];
//	if(WE_spsr) r_wcol <= DI[6];
	if(WE_spsr) r_spi2x <= DI[0];
      end
  end

endmodule
